//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//
// DUMM96 - Plzen Visibility Algorithms
// Idea by Robert F. Tobler & Dieter Schmalstieg
// Implementation by Dieter Schmalstieg
//
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

#include <graphics.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <assert.h>


//---------------------------------------------------------------------------
// definitions
//---------------------------------------------------------------------------


typedef int bool;
const int TRUE = 1;
const int FALSE = 0;

const int sc = 5;


struct Vec2D
{
    float x;
    float y;
    Vec2D() : x(0), y(0) {}
    Vec2D(float aX, float aY) : x(aX), y(aY) {}
};


struct Edge
{
    int tIdx;
    int eIdx;
    Edge() {}
    Edge(int atIdx, int aeIdx) : tIdx(atIdx), eIdx(aeIdx) {}
};

const int vertexNo = 100;
Vec2D v[vertexNo];


const int triangleNo = 200;
int t[triangleNo][3];
int n[triangleNo][3];
int w[triangleNo][3];


int vCount,tCount;

Vec2D eye;


//---------------------------------------------------------------------------
// system proprietary stuff
//---------------------------------------------------------------------------


void
graphicsMode()
{
    int gdriver = DETECT, gmode;
    initgraph(&gdriver, &gmode, "");
}


void
drawPoint(Vec2D point,int color)
{
   setcolor(color);
   circle(int(point.x*sc),int(point.y*sc),2);
}


void
drawLine(Vec2D v1,Vec2D v2,int color)
{
   setcolor(color);
   line(int(v1.x*sc),int(v1.y*sc),
	int(v2.x*sc),int(v2.y*sc));
}


int
navigation(Vec2D& eye)
{
    if(!kbhit()) return FALSE;
    char ch = getch();
    const float step = 2.0;
    switch(ch)
    {
	case '2': eye.y += step; break;
	case '4': eye.x -= step; break;
	case '6': eye.x += step; break;
	case '8': eye.y -= step; break;
	case 'q': exit(0);
    }
    return TRUE;
}


void
clearScreen()
{
    setbkcolor(0);
    clearviewport();
}

void fillTriangle(int i, int color)
{
    int p[6];
    setfillstyle(SOLID_FILL,color);

    for(int j=0; j<3; j++)
    {
	int p1 = t[i][j];
	p[j*2+0] = int(v[p1].x*sc);
	p[j*2+1] = int(v[p1].y*sc);
    }
    fillpoly(3,p);
}

//---------------------------------------------------------------------------
// aux functions
//---------------------------------------------------------------------------

void
drawEdge(Edge e, int color)
{
    drawLine(v[t[e.tIdx][e.eIdx]],v[t[e.tIdx][(e.eIdx+1)%3]],
    color);
}


bool
sameEdge(int v1, int v2, int u1, int u2)
{
    return (v1==u1 && v2==u2) || (v1==u2 && v2==u1);
}


int
commonEdge(int a, int b)
{
    for(int i=0; i<3; i++)
	for(int j=0; j<3; j++)
	    if(sameEdge(t[a][i],t[a][(i+1)%3],t[b][j],t[b][(j+1)%3]))
		return i;
    return -1;
}


Vec2D
centerOfGravity(int idx)
{
    Vec2D s;
    for(int i=0; i<3; i++)
    {
	s.x += v[t[idx][i]].x;
	s.y += v[t[idx][i]].y;
    }
    s.x /= 3;
    s.y /= 3;
    return s;
}


void drawTriangle(int i, int col1, int col2)
{
    int color;
    for(int j=0; j<3; j++)
    {
	if(w[i][j]) color = col1; else color = col2;
	setcolor(color);
	int p1 = t[i][j];
	int p2 = t[i][(j+1)%3];
	drawLine(v[p1],v[p2],color);
    }
}


bool
left(Vec2D testVec, Vec2D a, Vec2D b)
{
    Vec2D d1(testVec.x-a.x,testVec.y-a.y);
    Vec2D d2(b.x-a.x,b.y-a.y);

    float dot = d1.x*d2.y-d1.y*d2.x;
    return dot>0;
}


int
currentTriangle(Vec2D eye)
{
    for(int i=0; i<tCount; i++)
    {
	bool found = TRUE;
	for(int j=0; j<3; j++)
	    found = found && left(eye,v[t[i][j]],v[t[i][(j+1)%3]]);
	if(found) break;

    }
    if(i==tCount) return -1;
    return i;
}


void
drawTriangles()
{
    for(int i=0; i<tCount; i++) drawTriangle(i,15,1);
}


//---------------------------------------------------------------------------
// data structure building functions
//---------------------------------------------------------------------------

void
readWorld(char* fileName)
{
    FILE* f = fopen(fileName,"r");
    vCount=0;
    float x,y;
    do {
	fscanf(f,"%f,%f\n",&x,&y);
	if(x != -1) { v[vCount].x=x; v[vCount].y=y; }
	vCount++;
    } while(x != -1);

    tCount=0;
    int v1,v2,v3,t1,t2,t3;
    do {
	fscanf(f,"%d,%d,%d,%d,%d,%d\n",&v1,&v2,&v3,&t1,&t2,&t3);
	if(v1 != -1)
	{
	    t[tCount][0]=v1;
	    t[tCount][1]=v2;
	    t[tCount][2]=v3;
	    w[tCount][0]=t1;
	    w[tCount][1]=t2;
	    w[tCount][2]=t3;
	}
	tCount++;
    } while(v1 != -1);
    fclose(f);
}


void
computeNeighbors()
{
    int i,j;
    for(i=0; i<tCount; i++)
    {
	for(j=0; j<3; j++) n[i][j] = -1;
	int edge;
	for(j=0; j<tCount; j++)
	{
	    if(i != j && (edge=commonEdge(i,j)) != -1) n[i][edge] = j;
	}
    }
}


void
correctWalls()
{
    for(int i=0; i<tCount; i++)
	for(int j=0; j<3; j++)
	{
	    int neighbor = n[i][j];
	    if(neighbor != -1)
	    {
		int thisEdge = commonEdge(i,neighbor);
		int neighborEdge = commonEdge(neighbor,i);
		w[neighbor][neighborEdge] = w[i][thisEdge];
	    }
	    else
	    {
                w[i][j]=1;
	    }
	}
}

//---------------------------------------------------------------------------
// draw maze function: the new visibility computation function
//---------------------------------------------------------------------------


Vec2D
intersect(Vec2D a, Vec2D b, Vec2D c, Vec2D d, bool& ok)
{
    float d1x = b.x-a.x;
    float d1y = b.y-a.y;
    float d2x = d.x-c.x;
    float d2y = d.y-c.y;
    float n = (d1x*d2y-d1y*d2x);
    ok = (n != 0.0);

    if(ok)
    {
	float t = (a.y*d2x-c.y*d2x-a.x*d2y+c.x*d2y)/n;
	return Vec2D(a.x+t*d1x,a.y+t*d1y);
    }
    else return Vec2D(0,0);
}


bool
isWall(Edge e)
{
    return w[e.tIdx][e.eIdx];
}

void
drawColumn(Vec2D vec)
{
    drawPoint(vec,13);
}


Vec2D
computeDelta(Edge e)
{
    Vec2D v2 = v[t[e.tIdx][e.eIdx]];
    Vec2D v1 = v[t[e.tIdx][(e.eIdx+1)%3]];
    const float deltaFactor = 10;
    Vec2D d((v2.x-v1.x)/deltaFactor,(v2.y-v1.y)/deltaFactor);
    return d;
}


void
vecAdd(Vec2D& v, Vec2D delta)
{
    v.x += delta.x;
    v.y += delta.y;
}


Edge
getNewLeftEdge(Edge e)
{
    int neighbor = n[e.tIdx][e.eIdx];
    int neighborEdge = commonEdge(neighbor,e.tIdx);
    int leftEdge = (neighborEdge+2)%3;

    return Edge(neighbor,leftEdge);
}

Edge
getNewRightEdge(Edge e)
{
    int neighbor = n[e.tIdx][e.eIdx];
    int neighborEdge = commonEdge(neighbor,e.tIdx);
    int rightEdge = (neighborEdge+1)%3;

    return Edge(neighbor,rightEdge);
}


Vec2D currentVec;

void
drawMaze(Edge front, Vec2D clipVec)
{

    fillTriangle(front.tIdx,8);
    drawTriangle(front.tIdx,4,5);

    if(isWall(front))
    {
	Vec2D delta = computeDelta(front);
	Vec2D rightVec  = v[t[front.tIdx][front.eIdx]];
	Vec2D leftVec = v[t[front.tIdx][(front.eIdx+1)%3]];

	bool ok; Vec2D projVec;
	projVec = intersect(eye,currentVec,leftVec,rightVec, ok);
	if(ok) currentVec = projVec;

	while(left(currentVec,eye,clipVec))
	{
	    drawColumn(currentVec);
	    vecAdd(currentVec,delta);
	}
	return;
    }
    Edge le = getNewLeftEdge(front);
    Edge re = getNewRightEdge(front);
    Vec2D newVec = v[t[le.tIdx][le.eIdx]];

    if(left(newVec,eye,currentVec))
    {
	drawMaze(re,clipVec);
    }
    else if(left(newVec,eye,clipVec))
    {
	drawMaze(le,newVec);
	drawMaze(re,clipVec);
    }
    else
    {
	drawMaze(le,clipVec);
    }
}


//---------------------------------------------------------------------------
// main program
//---------------------------------------------------------------------------

void
main()
{
    int i,j,color;
    eye = Vec2D(15,55);

    printf("reading world...\n");
    readWorld("world.txt");

    printf("finding neighbors...\n");
    computeNeighbors();

    printf("correcting walls...\n");
    correctWalls();


    graphicsMode();


    while(1)
    {
	clearScreen();
	drawTriangles();
	int ct = currentTriangle(eye);
	if(ct != -1)
	{
	    drawTriangle(ct,4,5);

	    for(int i=0; i<3; i++)
	    {
		Vec2D clipVec = v[t[ct][i]];
		currentVec = v[t[ct][(i+1)%3]];
		drawMaze(Edge(ct,i),clipVec);
	    }
	}
	drawPoint(eye,15);
	while(!navigation(eye));
    }

}
